/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.util.*;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import net.hoopajoo.swingx.*;
import java.io.*;

// The main mapper panel
//class CheeseCanvas extends JPanel implements MouseListener,MouseMotionListener,KeyListener {
class CheeseCanvas extends JComponent implements MouseListener,MouseMotionListener,KeyListener {
	final static int TOOL_SELECT = 0;
	final static int TOOL_MERGE = 1;
	final static int TOOL_MOVE = 2;
	final static int TOOL_DELETE = 3;
	final static int TOOL_POINT = 4;
	final String [] toolnames = {
			"select_tool",
			"merge_tool",
			"move_tool",
			"delete_tool",
			"point_tool"
	};
	Cursor [] tool_cursors = {
			new Cursor( Cursor.DEFAULT_CURSOR ),
			new Cursor( Cursor.HAND_CURSOR ),
			new Cursor( Cursor.MOVE_CURSOR ),
			createCursor("delete"),
			new Cursor( Cursor.CROSSHAIR_CURSOR )
	};
	
	private CheeseMapData map;
	int selectedid = 0;
	int tool_active = TOOL_SELECT;
	String tool_saved = "tool_select";
	int ctx_edit_roomid = 0;
	Point3d drag_start;
	Point drag_start2d, drag_end2d;
	boolean lclick_dragging = false;
	boolean rclick_dragging = false;
	public PointToolCallback point_callback = null;
	Cursor cursor_saved = getCursor();
	Cursor cursor_movemap = new Cursor( Cursor.MOVE_CURSOR );
	Cursor cursor_busy = new Cursor( Cursor.WAIT_CURSOR );

	// Gridded overlays, TODO: make this a gui option
	boolean render_grid = true;
	
	// futures are buggy
	boolean render_future = false;
	
	// TODO: These have no ui yet
	boolean render_notes = true;
	int render_notes_length = 15;
	
		String timer_section;
	Date timer_start, timer_end;
	
	HashMap selected_rooms = new HashMap();
	
	int tile_width = 32,
		tile_height = 32,
		origin_x, origin_y, origin_z,
		rows = 0, cols = 0;
	private String theme;
	private long changed = 0,
		last_sync = 0;
	
	//private VolatileImage screencache = null;
	//private BufferedImage screencache = null;
	private BufferedImage screencache = null;
		
	
	// TODO: Needs another map for temp overlays, that can be image/text
	// so that we can "blink" people off where.  These overlays
	// should have an expire time (+5 seconds?) that will make them
	// auto-delete at the next render after the expiry time
	// this would also make the "hud" display of hp/move/mana possible
	
	private HashMap mapdata = new HashMap();
	private HashMap mtimedata = new HashMap();
	private HashMap tilenames = new HashMap();
	private Vector warplines = new Vector();

	// HashMap has faster get but slower put apparently, but it'll work
	private HashMap image_cache = new HashMap();

	CheeseCanvas() {
		addMouseListener( this );
		addMouseMotionListener( this );
		addKeyListener( this );
		setToolTipText( "Room Info" );

		// TODO: Load this from a file
		// Set tile mapping
		tilenames.put("@","indoors");
		tilenames.put("<","mountain");
		tilenames.put("+","road");
		tilenames.put("#","city");
		tilenames.put("~","shallow_water");
		tilenames.put("\"","brush");
		tilenames.put("_","field");
		tilenames.put("w","water");
		tilenames.put("=","tunnel");
		tilenames.put("%","forest");
		tilenames.put("W","rapids");
		tilenames.put("O","cavern");
		tilenames.put("(","hills");
		tilenames.put("U","underwater");
		
		setDoubleBuffered( false );
		setFocusable( true );
		// requestDefaultFocus();
		// requestFocus();
		requestFocusInWindow();
	}

	public void update_map( byte[] buffer ) {
		mapdata.clear();
		mtimedata.clear();
		
		String s = new String( buffer );
		BufferedReader in = new BufferedReader( new StringReader( s ) );
		String line;
		try{
			while( (line = in.readLine()) != null ) {
				String[] values = line.split( "," );

				//CheeseMap.dmesg( "Got map room: " + values[ 0 ] + " mtime: " + values[ 4 ] );
				mapdata.put( values[ 1 ] + "," + values[ 2 ] + "," + values[ 3 ], 
						new Integer( values[ 0 ] ) );
				mtimedata.put( values[ 1 ] + "," + values[ 2 ] + "," + values[ 3 ], 
						new Long( values[ 4 ] ) );
			}
		}catch( Exception e ){
			e.printStackTrace();
		}		
	}
	
	public void mouseMoved(MouseEvent e) {
	}
	
	public void mouseDragged(MouseEvent e) {
		if( lclick_dragging && tool_active == TOOL_SELECT ) {
			// draggin to select
			drag_end2d = e.getPoint();
			repaint();
		}
		if( rclick_dragging ) {
			// dragging the map around
			drag_end2d = e.getPoint();
			repaint();
		}
	}
	
	public void mousePressed(MouseEvent e) {
		java.awt.Point p = e.getPoint();
		
		if( e.getButton() == 1 ) {
			CheeseLab.dmesg( 3, "Button 1 down" );

			drag_start = local_to_global(p.x, p.y);
			drag_start2d = p;
			drag_end2d = p; // ends here too atm
			lclick_dragging = true;
			
		}else if( e.getButton() == 3 ) {
			CheeseLab.dmesg( 3, "Button 3 down" );
			drag_start = local_to_global( p.x, p.y );
			drag_start2d = p;
			drag_end2d = p; // ends here too atm
			rclick_dragging = true;
			cursor_saved = getCursor();
			setCursor( cursor_movemap );
		}
	}

	public void mouseReleased(MouseEvent e) {
		java.awt.Point p = e.getPoint();			
		Point3d drag_end = local_to_global( p.x, p.y );

		if( e.getButton() == 1 ) {
			if( tool_active == TOOL_SELECT ) {
				boolean select_3d = false;
				
				if(!e.isShiftDown()) 	// Allow adding rooms to selection by holding the SHIFT key
					selected_rooms.clear();
				
				if(e.isControlDown())	// Do 3D select if Control was held down
					select_3d = true;
				
				CheeseLab.dmesg( 3, "Button 1 up" );
				
				int x, y, z,
					min_x, max_x,
					min_y, max_y,
					min_z, max_z;
				
				min_z = max_z = drag_start.z;
				
				if( drag_start.x < drag_end.x ) {
					min_x = drag_start.x;
					max_x = drag_end.x;
				}else{
					min_x = drag_end.x;
					max_x = drag_start.x;
				}
				
				if( drag_start.y < drag_end.y ) {
					min_y = drag_start.y;
					max_y = drag_end.y;
				}else{
					min_y = drag_end.y;
					max_y = drag_start.y;
				}
				
				/* TODO: Actually implement 3D select. We need to make selection algorithm more
				 *       efficent. One way is to implement search by coordinates in broker.
				 */
				for( z = min_z; z <= max_z; z++ ) {
					CheeseLab.dmesg( 3, "Selecting rooms from: " + min_x + "," + min_y + "," + z +
							" " + max_x + "," + max_y + "," + z );
					for( x = min_x; x <= max_x; x++ ) {
						for( y = min_y; y <= max_y; y++ ) {
							// These rooms ought to be in our map since you can't really
							// click on them unless you can see them
							String global_key = new String( x + "," + y + "," + z );
							Integer roomid = (Integer)mapdata.get( global_key );
							if( roomid != null ) {
								CheeseLab.dmesg( 3, "Selecting: " + roomid );
								selected_rooms.put( roomid, roomid );
							}
						}
					}
				}
				
				/* TODO: Maybe we should move map if there was only one room selected.
				 *       The need for LMB navigation is debatable though
				 * 
				 * if( selected_rooms.size() == 1 ) {
				 *	Point3d p3d = local_to_global( p.x, p.y );
				 *	center_on( p3d.x, p3d.y, p3d.z );
				 *	CheeseMap.set_follow_player( false );
				 * }
				 */
				
				requestFocus();
			}else if( tool_active == TOOL_DELETE ) {
				// delete
				String global_key = new String( drag_end.x + "," + drag_end.y + "," + drag_end.z );
				Integer roomid = (Integer)mapdata.get( global_key );
				if( roomid != null )
					CheeseMap.send_tx( 1, "delete", roomid.toString() );
			}else if( tool_active == TOOL_MERGE ) {
				// merge
				String global_key = new String( drag_end.x + "," + drag_end.y + "," + drag_end.z );
				Integer roomid = (Integer)mapdata.get( global_key );
				if( roomid != null )
					CheeseMap.send_tx( 1, "merge", roomid.toString() );
			}else if( tool_active == TOOL_POINT ) {
				// Restore last tool. Execute callback.
				setEditTool( tool_saved );
				if (point_callback != null) {
					point_callback.pointedTo(p);
					point_callback = null;
				}
			}
			
			lclick_dragging = false;
			changed();
		}
		
		if( e.getButton() == 3 ) {
			int dx = drag_start.x - drag_end.x;
			int dy = drag_start.y - drag_end.y;

			rclick_dragging = false;
			setCursor( cursor_saved );

			if( dx != 0 || dy != 0 ) {
				move_origin( dx, dy, 0 );
				CheeseMap.set_follow_player( false );
			}else{
				// Call repaint to cancel visual effect of "accidental" dragging
				repaint();
				
				// Now build and open context menu
				JPopupMenu menu = new JPopupMenu();
				JMenuItem item;
				CheeseRoomData room = null;

				// Context menu room editor
				String r_name;
				Integer roomid = (Integer)mapdata.get(drag_end.x + ","
						+ drag_end.y + "," + drag_end.z);
				if( roomid != null ) {
					room = map.get_room( roomid.intValue() );
					r_name = room.name;
					ctx_edit_roomid = roomid.intValue();
				}else{
					r_name = "No Room Here";
					ctx_edit_roomid = 0;
				}

				// Trim room name length to show it in context menu
				if( r_name.length() > 25 )
					r_name = r_name.substring( 0, 22 ) + "... ";

				item = new JMenuItem( r_name );
				item.setEnabled( false );
				menu.add( item );
				menu.addSeparator();

				if( ctx_edit_roomid != 0 ) {
					item = new JMenuItem( "Open Room Editor" );
					item.addActionListener( new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							selectedid = ctx_edit_roomid;
							CheeseLab.dmesg(3, "Do editor for: " + ctx_edit_roomid);
							CheeseMap.set_edit_room(map.get_room(ctx_edit_roomid));
						}
					} );
					menu.add(item);
					
					// Construct sub-menu for Overlay editing
					String [] ovrl = CheeseMapRoomEditor.common_overlays;
					JMenu submenu = new JMenu( "Overlays" );
					for( int i = 0; i < ovrl.length; i++) {
					    boolean checked = false; 
					    ImageIcon icon = new ImageIcon(
					            getImage( "overlay_" + ovrl[i] ).
					            	getScaledInstance( 16, 16,
					            	        Image.SCALE_SMOOTH ) );
					    for( int j = 0; j < room.overlays.length; j++ ) {
					        if( room.overlays[j].equals( ovrl[i] ) )
					            checked = true;
					    }
					    item = new JCheckBoxMenuItem( ovrl[i], icon, checked );
					    item.addActionListener( new CtxEditorListener( 
					            		ovrl[i], ctx_edit_roomid ) );
					    item.setActionCommand( "set_overlay" );
					    submenu.add( item );
					}
					menu.add( submenu );

					// Construct sub-menus for exit editing
					String[] dirs = { "N", "S", "E", "W", "U", "D" };
					int[] cur_exits = { room.exit_n, room.exit_s, room.exit_e,
							room.exit_w, room.exit_u, room.exit_d };

					for( int i = 0; i < dirs.length; i++ ) {
						JMenu m = new JMenu( "Exit " + dirs[i] );
						ButtonGroup bg = new ButtonGroup();

						bg.add( item = new JRadioButtonMenuItem( "None" ) );
						if( cur_exits[i] == 0 )
							item.setSelected( true );
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_none" );
						m.add( item );
						bg.add( item = new JRadioButtonMenuItem( "Unexplored" ) );
						if( cur_exits[i] == -1 )
							item.setSelected( true );
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_unexplored" );
						m.add( item );
						bg.add( item = new JRadioButtonMenuItem( "Death Trap" ) );
						if( cur_exits[i] == -2 )
							item.setSelected( true );
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_dt" );
						m.add( item );
						bg.add( item = new JRadioButtonMenuItem( "To room ..." ) );
						if( cur_exits[i] > 0 ) {
							item.setText( "To room " + cur_exits[i] );
							item.setSelected( true );
						}
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_from_map" );
						m.add( item );

						menu.add( m );
					}
				}

				// Main operations
				menu.addSeparator();
				menu.add( new JSelectedCheckMenuItem( CheeseMap.follow_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.levelup_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.leveldown_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.zoomin_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.zoomout_action ) );
				
				// Toolbox
				menu.add( CheeseMap.toolbox.createMenu( "Tools" ) );

				// Phew! Done! Show it.
				// menu.setLightWeightPopupEnabled(false);
				menu.show( this, p.x, p.y );
			}
		}
	}
	
	public void mouseEntered(MouseEvent e) { }
	public void mouseExited(MouseEvent e) { }
	public void mouseClicked(MouseEvent e) { }

	public String getToolTipText( MouseEvent e ) {
		java.awt.Point p = e.getPoint();

		Point3d p3d = local_to_global( p.x, p.y );
		
		// TODO: make this useful :p
		String ret = new String();
		Integer roomid = (Integer)mapdata.get( p3d.x + "," + p3d.y + "," + p3d.z );
		Long mtime = (Long)mtimedata.get( p3d.x + "," + p3d.y + "," + p3d.z );
		if( roomid != null ) {
			CheeseRoomData room = map.get_room( roomid.intValue(), mtime.longValue() );
			String show_notes = room.notes;
			if( show_notes == null )
				show_notes = "";
			
			ret = "<html>" + 
				"#" + room.roomid + " - " + room.x + "," + room.y + "," + room.z + "<br>" + 
				room.name + 
				"<br>" +
				show_notes +
				"</html>";
		}else{
			ret = "No Room Here";
		}

		return( ret );
	}

	public void paintComponent(Graphics g) {
		if( CheeseMap.online == false ) {
			super.paintComponent( g );
			return;
		}
		
		// Handle right-click map move mode
		if( rclick_dragging ) {
			if( screencache != null ) // We can't move around nonexisting picture right?
				g.drawImage( screencache, drag_end2d.x - drag_start2d.x,
						drag_end2d.y - drag_start2d.y, null );
			return;
		}
		
		// Handle resize refreshes
		// TODO: take player position into account if in follow_player mode
		if( screencache != null ) {
			// validate width/height
			if( screencache.getWidth(null) != getWidth() ) {
				screencache.flush();
				screencache = null;
			}else if( screencache.getHeight(null) != getHeight() ) {
				screencache.flush();
				screencache = null;
			}
		}

		// If no screencache, then we are definately gonna need to repaint
		if( screencache == null )
			changed( false );
		
		// Paint to a bitmap, that way exposed paint events don't flicker like mad
		if( changed >= last_sync ) {
			// We mark the sync now so that if there are changed while we
			// are rendering, we will get them
			int rend_x = origin_x,
				rend_y = origin_y,
				rend_z = origin_z;
			last_sync = System.currentTimeMillis();

			CheeseLab.dmesg( 2, "Redraw!" );
			cursor_saved = getCursor();
			setCursor( cursor_busy );

			Date start = new Date();
			
			cols = getWidth() / tile_width;
			rows = getHeight() / tile_height;

			CheeseLab.dmesg( 3, "Size: " + rows + "," + cols );
			CheeseLab.dmesg( 3, "Origin: " + rend_x + "," + rend_y + "," + rend_z );
			
			// Figure our bounding box
			int start_x = rend_x;
			int start_y = rend_y;
			int start_z = rend_z;

			int end_x = start_x + cols;
			int end_y = start_y + rows;
			int end_z = rend_z + 1;

			// Request an update from the server (syncro though, this opens up a lotta race conditions...)
			CheeseMap.clearping( "map" );
			CheeseMap.send_tx( 1, "map", 
				start_x + "," + start_y + "," + start_z + " " +
				end_x + "," + end_y + "," + end_z );
			CheeseMap.send_tx( 1, "ping", "map" );
			while( ! CheeseMap.checkping() && CheeseMap.online ) {
				try {
					CheeseLab.dmesg( 3, "Waiting for map to complete" );
					Thread.sleep( 100 );
				}catch( Exception e ) {
					e.printStackTrace();
				}
			}
			// Now out map data is updated in mapdata
			
			// Freeze players
			CheesePlayerData[] players = map.get_players();


			if( screencache == null ) {
				// make cache object
				//screencache = createImage( getWidth(), getHeight() );
				CheeseLab.dmesg( 2, "Creating new buffered image for screen cache" );
				screencache = (BufferedImage)createImage( getWidth(), getHeight() );
			}

			warplines.clear();
			Graphics2D sg = (Graphics2D)screencache.getGraphics();
			
			// Make a slightly larger font
			sg.setFont( new Font( null, Font.BOLD, 12 ) );
			
			sg.setColor( new Color( 50, 50, 50 ) );
			sg.fillRect( 0, 0, getWidth(), getHeight() );
			
			Vector textnodes = new Vector();
			int x, y, z, local_x, local_y;
			for( z = 1; z >= 0; z-- ) {	
				for( x = 0; x <= cols; x++ ) {
					local_x = x * tile_width;
					for( y = 0; y <= rows; y++ ) {
						local_y = y * tile_height;
						//System.out.println( "draw image: " + local_x + "," + local_y );
						int global_x = start_x + x;
						int global_y = start_y + y;
						int global_z = start_z + z;

						String global_key = new String( global_x + "," + global_y + "," + global_z );
						Integer roomid = (Integer)mapdata.get( global_key );
						Long mtime = (Long)mtimedata.get( global_key );
						if( roomid != null ) {
							CheeseRoomData room = map.get_room( roomid.intValue(), mtime.longValue() );
							String tilename = (String)tilenames.get( room.terrain );
							if( tilename == null )
								tilename = "unknown";

							sg.drawImage( getImage( "t_" + tilename, render_grid ), local_x, local_y, null );
							//sg.drawImage( getImage( "grid" ), local_x, local_y, null );

							String exittile;

							exittile = get_exit_tile( room, room.exit_e, room.door_e, "e" );
							if( exittile != null ) sg.drawImage( getImage( exittile ), local_x, local_y, null );
							exittile = get_exit_tile( room, room.exit_w, room.door_w, "w" );
							if( exittile != null ) sg.drawImage( getImage( exittile ), local_x, local_y, null );
							exittile = get_exit_tile( room, room.exit_n, room.door_n, "n" );
							if( exittile != null ) sg.drawImage( getImage( exittile ), local_x, local_y, null );
							exittile = get_exit_tile( room, room.exit_s, room.door_s, "s" );
							if( exittile != null ) sg.drawImage( getImage( exittile ), local_x, local_y, null );
							exittile = get_exit_tile( room, room.exit_u, room.door_u, "u" );
							if( exittile != null ) sg.drawImage( getImage( exittile ), local_x, local_y, null );
							exittile = get_exit_tile( room, room.exit_d, room.door_d, "d" );
							if( exittile != null ) sg.drawImage( getImage( exittile ), local_x, local_y, null );
							
							if( global_z == rend_z ) {
								boolean drew_an_overlay = false;
								for(int i = 0; i < room.overlays.length; i++ ) {
									if( ! room.overlays[ i ] .equals( "" ) ) {
										CheeseLab.dmesg( 3, "Overlay: " + room.overlays[ i ] );
										sg.drawImage( getImage( "overlay_" + room.overlays[ i ] ), local_x, local_y, null );
										drew_an_overlay = true;
									}
								}
								
								// render notes text
								if( room.notes != null ) {
									if( room.notes.length() < render_notes_length && render_notes) {
										textnodes.add( room );
									}else{
										if( ! drew_an_overlay ) {
											// Render a note tile if notes are here
											CheeseLab.dmesg( 3, "Doing default notes overlay" );
											sg.drawImage( getImage( "overlay_notes" ), local_x, local_y, null );
										}
									}
								}
								
							}else{
								sg.drawImage( getImage( "overlay_blacken" ), local_x, local_y, null );
							}
							
							// Selected
							if( selected_rooms.get( roomid ) != null ) {
								CheeseLab.dmesg( 3, "Drawing selected" );
								sg.drawImage( getImage( "selected" ), local_x, local_y, null );   
							}
							
							if( room.roomid == CheeseMap.edit_roomid )
								sg.drawImage( getImage( "infobox" ), local_x, local_y, null );
						}
					}
				}
			}

			// warp lines
			sg.setColor( new Color( 255, 0, 0 ) );
			for( int i = 0; i < warplines.size(); i++ ) {
				int[] points = (int[])warplines.get( i );
				Point pstart = global_to_local_reference( points[ 0 ], points[ 1 ], points[ 2 ],
						rend_x, rend_y, rend_z );
				Point pend = global_to_local_reference( points[ 3 ], points[ 4 ], points[ 5 ],
						rend_x, rend_y, rend_z );
				
				sg.drawLine( pstart.x + points [ 6 ], pstart.y + points[ 7 ],
					pend.x + tile_width / 2, pend.y + tile_height / 2 );
			}
			
			// Notes
			for( int i = 0; i < textnodes.size(); i++ ) {
				CheeseRoomData room =(CheeseRoomData)textnodes.get( i );
				Point p = global_to_local_reference( room.x, room.y, room.z, rend_x, rend_y, rend_z );
				local_x = p.x;
				local_y = p.y;
				
				int offset_x = 4;
				int offset_y = tile_height - 4;
				
				if( room.x % 2 == 0 ) {
					offset_y = tile_height / 2;
				}
				
				sg.setColor( Color.BLACK );
				sg.drawString( room.notes, 
					local_x + offset_x, local_y + offset_y );
				sg.setColor( Color.LIGHT_GRAY );
				sg.drawString( room.notes, 
					local_x + offset_x + 1, local_y + offset_y + 1 );
			}
			
			// Player overlays
			CheeseLab.dmesg( 3, "Loading player map" );
			int hud_y = 5,
				hud_x = getWidth() - 5;
			sg.setColor( Color.WHITE );
			FontMetrics metrics = sg.getFontMetrics();
			for( int i = 0; i < players.length; i++ ) {
				Point p = global_to_local_reference( players[ i ].x, players[ i ].y, players[ i ].z, rend_x, rend_y, rend_z );
				
				if( players[ i ].z == rend_z ) {
					if( p.x >= 0 && p.x <= getWidth() && p.y >= 0 && p.y <= getHeight() ) {
						String playername = players[ i ].name;
						sg.drawImage( getImage( "player_" + playername ), p.x, p.y, null );
						
						if( players[ i ].roomid == 0 )
							sg.drawImage( getImage( "player_lost" ), p.x, p.y, null );						
					}
				}
				
				// Do HUD
				if( players[ i ].online ) { // TODO: add selective showing later
					String username = players[ i ].name;
					if( username.equals( CheeseMap.follow_userid ) ) {
						username = "YOU";
					}
					
					String player_info_string = username + ": " +
						players[ i ].hit_curr + "/" + players[ i ].hit_max + " hits " +
						players[ i ].mana_curr + "/" + players[ i ].mana_max + " mana " +
						players[ i ].move_curr + "/" + players[ i ].move_max + " moves";
						
					Rectangle2D bounds = metrics.getStringBounds( player_info_string, sg );
					sg.setColor( Color.BLACK );
					sg.drawString( player_info_string, 
						(int)(hud_x - bounds.getWidth()), (int)(hud_y + bounds.getHeight()) );
					sg.setColor( Color.WHITE );
					sg.drawString( player_info_string, 
						(int)(hud_x - bounds.getWidth()) + 1, (int)(hud_y + bounds.getHeight()) + 1 );
					
					hud_y += bounds.getHeight();
					
					if( players[ i ].name.equals( CheeseMap.follow_userid ) && render_future ) {
						// Draw where we are going to
						CheeseLab.dmesg( 3, "Drawing future placement" );
						Point last_point = global_to_local_reference( players[ i ].x, players[ i ].y, players[ i ].z,
								rend_x, rend_y, rend_z );
						last_point.x += (int)(tile_width / 2);
						last_point.y += (int)(tile_height / 2);
						sg.setColor( Color.YELLOW );
						for( int ii = 0; ii < players[ i ].future.length; ii++ ) {
							if( ! players[ i ].future[ ii ].equals( "" ) ) {
								String[] split = players[ i ].future[ ii ].split( "," );
								Point next_point = global_to_local_reference( 
										Integer.parseInt( split[ 1 ] ), 
										Integer.parseInt( split[ 2 ] ), 
										Integer.parseInt( split[ 3 ] ),
										rend_x, rend_y, rend_z ); 
								next_point.x += (int)(tile_width / 2);
								next_point.y += (int)(tile_height / 2);
								
								sg.drawLine( last_point.x, last_point.y, next_point.x, next_point.y );
								last_point = next_point;
							}
						}
					}
				}
			}

			Date end = new Date();
			long diff = end.getTime() - start.getTime();
			double secs = (double)diff / 1000;
			CheeseLab.dmesg( 1, "Took: " + diff + "ms (" + secs + " secs)" );
			
			sg.dispose();
			
			setCursor( cursor_saved );
		}

		g.drawImage( screencache, 0, 0, null );
		
		// if dragging draw that
		if( lclick_dragging ) {
			CheeseLab.dmesg( 3, "Drawing rect" );
			g.setColor( Color.RED );
			g.drawRect( drag_start2d.x, drag_start2d.y,
					drag_end2d.x - drag_start2d.x,
					drag_end2d.y - drag_start2d.y );
		}
	}

	private String get_exit_tile( CheeseRoomData room, int exit, String door, String dir ) {
		String tilename = null;
		
		if( exit == 0 ) {
			tilename = "overlay_noexit_" + dir;
		}else if( exit == -2 ) {
			tilename = "overlay_dt_" + dir;
		}else if( exit == -1 ) {
			tilename = "overlay_unexplored_" + dir;
		}else {
			if( ! door.equals( "" ) ) {
				if( door.startsWith( "-" ) ) {
					tilename = "overlay_door_unlocked_" + dir;
				}else if( door.startsWith( "+" ) ) {
					tilename = "overlay_door_locked_" + dir;
				}else{
					tilename = "overlay_door_" + dir;
				}
			}else if( dir.equals( "u" ) || dir.equals( "d" ) ) {
				tilename = "overlay_exit_" + dir;
			}

			// Check for warping
			CheeseRoomData connect = map.get_room( exit );

			// calc diff from this room to that room, if it's more than 1
			// than we need to render warp lines
			int roomdiff = 0;
			
			roomdiff += Math.abs( connect.x - room.x );
			roomdiff += Math.abs( connect.y - room.y );
			roomdiff += Math.abs( connect.z - room.z );

			if( roomdiff == 1 ) {
				// Make sure it's in the right direction
				if( dir.equals( "e" ) ) {
					if( room.x - connect.x != -1 )
						roomdiff = 0;
				}else if( dir.equals( "w" ) ) {
					if( room.x - connect.x != 1 )
						roomdiff = 0;
				}else if( dir.equals( "n" ) ) {
					if( room.y - connect.y != 1 )
						roomdiff = 0;
				}else if( dir.equals( "s" ) ) {
					if( room.y - connect.y != -1 )
						roomdiff = 0;
				}else if( dir.equals( "u" ) ) {
					if( room.z - connect.z != -1 )
						roomdiff = 0;
				}else if( dir.equals( "d" ) ) {
					if( room.z - connect.z != 1 )
						roomdiff = 0;
				}
			}

			if( roomdiff != 1 ) {
				//CheeseMap.dmesg( "Evaluating warp lines" );
				if( connect.z == room.z ) {
					// Draw warp lines
					int offset_x = 0;
					int offset_y = 0;
					if( dir.equals( "e" ) ) {
						offset_x = tile_width;
						offset_y = tile_height / 2;
					}else if( dir.equals( "w" ) ) {
						offset_x = 0;
						offset_y = tile_height / 2;
					}else if( dir.equals( "n" ) ) {
						offset_x = tile_width / 2;
						offset_y = 0;
					}else if( dir.equals( "s" ) ) {
						offset_x = tile_width / 2;
						offset_y = tile_height;
					}else if( dir.equals( "u" ) ) {
						offset_x = 3;
						offset_y = 3;
					}else if( dir.equals( "d" ) ) {
						offset_x = tile_width - 3;
						offset_y = tile_height - 3;
					}

					CheeseLab.dmesg( 3, "Adding warplines" );
					
					int[] line = new int[ 8 ];
					line[ 0 ] = room.x;
					line[ 1 ] = room.y;
					line[ 2 ] = room.z;

					line[ 3 ] = connect.x;
					line[ 4 ] = connect.y;
					line[ 5 ] = connect.z;
					
					line[ 6 ] = offset_x;
					line[ 7 ] = offset_y;

					warplines.add( line );
				}else{
					// Change tile to "overlay_warp"
					// because of diff z planes (connect is not visible)
					tilename = "overlay_warp_" + dir;
				}
			}
		}

		return( tilename );
	}

	Image getImage( String name ) {
		return( getImage( name, false ) );
	}
	
	Image getImage( String name, boolean gridded ) {
		//System.out.println( "Getting tile: " + name );

		// Load an image
		// 1) Request and load from server first, then save to file
		// 2) Load from disk, search in local/theme, local, theme/

		// TODO: request from server? or no server-side tiles?

		// If not in cache load it
		//Image im = (Image)image_cache.get( name );
		Image im = (Image)image_cache.get( name );
		//System.out.println( "Got ptr: " + im );
		if( im == null ) {
			CheeseLab.dmesg( 3, "Image not in cache, getting" );
			Toolkit toolkit = Toolkit.getDefaultToolkit();
			
			// TODO: Try file in tiles/name.png then cacheprefix/tiles/name.png
			String tilenames[] = { 
				CheeseMap.cachedir + "/tiles/" + name + ".png",
				"tiles/" + theme + "/" + name + ".png",
				"tiles/default/" + name + ".png",
				"tiles/" + name + ".png"
			};

			CheeseLab.dmesg( 3, "Loading: " + name );
			for( int i = 0; i < tilenames.length; i++ ) {
				ImageIcon ii = new ImageIcon( tilenames[ i ] );
				if( ii.getIconWidth() != -1 ) {
					CheeseLab.dmesg( 3, "Got match: " + tilenames[ i ] );
					im = ii.getImage().getScaledInstance( tile_width, tile_height, Image.SCALE_SMOOTH );
					
					// Force loading
					while( ! prepareImage( im, -1, -1, this ) );
			
					// Now convert to a compatible image from this gc
					Graphics2D sg = screencache.createGraphics();
					
					// Alpha blending does not do acceleration, but I like it, and under non-win32 environments the accel doesn't work anyways :(
					// so you all have to suffer!! *hop*
					Image t = sg.getDeviceConfiguration().createCompatibleImage( tile_width, tile_height, Transparency.BITMASK );
					//Image t = sg.getDeviceConfiguration().createCompatibleImage( tile_width, tile_height, Transparency.TRANSLUCENT );
					Graphics gg = t.getGraphics();
					gg.drawImage( im, 0, 0, null );
					
					if( gridded )
						gg.drawImage( getImage( "grid" ), 0, 0, null );
						
					gg.dispose();
					sg.dispose();

					im = t;
				}
			}
			
			// Cache it in memory if we got one
			if( im != null ) {
				image_cache.put( name, im );
			}
		}

		// if we still don't have an image, return missing
		if( im == null )
			return( getImage( "missing" ) );

		return( im );
	}

	public Point global_to_local_reference( int x, int y, int z, int rx, int ry, int rxz ) {
		Point p = new Point();
		p.x = ( x - rx ) * tile_width;
		//p.y = ( rows - ( y - origin_y ) ) * tile_height;
		p.y = ( y - ry ) * tile_height;
		return( p );
	}
	
	public Point global_to_local( int x, int y, int z ) {
		return( global_to_local_reference( x, y, z, origin_x, origin_y, origin_z ) );
	}
	
	public Point3d local_to_global( int x, int y ) {
		Point3d p = new Point3d();
		
		p.x = origin_x + ( x / tile_width );
		//p.y = origin_y + ( rows - ( y / tile_width ) );
		p.y = origin_y + ( y / tile_height );
		p.z = origin_z;
		
		return( p );
	}
	
	public void set_map( CheeseMapData m ) {
		map = m;
	}

	public void center_on( int x, int y, int z ) {
		CheeseLab.dmesg( 2, "Centering on: " + x + "," + y + "," + z );
		set_origin( x - ( cols / 2 ), y - ( rows / 2 ), z );
	}
	
	public void move_origin( int dx, int dy, int dz ) {
		CheeseLab.dmesg( 2, "Moving: " + dx + "," + dy + "," + dz );
		set_origin( origin_x + dx, origin_y + dy, origin_z + dz );
	}
	
	public void set_origin( int x, int y, int z ) {
		CheeseLab.dmesg( 2, "Setting origin: " + x + "," + y + "," + z );
		origin_x = x;
		origin_y = y;
		origin_z = z;

		changed();
	}

	public void changed() {
		changed( true );
	}
	
	public void changed( boolean redraw ) {
		changed = System.currentTimeMillis();
		if( redraw )
			repaint();
	}

	public void keyTyped( KeyEvent e ) {	}
	public void keyReleased( KeyEvent e ) { }

	public void keyPressed( KeyEvent e ) {
		// If we have selected items
		if( ! selected_rooms.isEmpty() ) {
			int x = 0,
				y = 0,
				z = 0;
			int delta = 1;
			
			if( (e.getModifiers() & InputEvent.SHIFT_MASK) != 0 ) {
				delta = 5;
			}
			
			if( e.getKeyCode() == KeyEvent.VK_UP ) {
				y -= delta;
			}else if( e.getKeyCode() == KeyEvent.VK_DOWN ) {
				y += delta;
			}else if( e.getKeyCode() == KeyEvent.VK_LEFT ) {
				x -= delta;
			}else if( e.getKeyCode() == KeyEvent.VK_RIGHT ) {
				x += delta;
			}else if( e.getKeyCode() == KeyEvent.VK_PAGE_UP ) {
				z++;
			}else if( e.getKeyCode() == KeyEvent.VK_PAGE_DOWN ) {
				z--;
			}
			
			if( x != 0 || y != 0 || z != 0 ) {
				CheeseLab.dmesg( 3, "Gonna move rooms: " + x + "," + y + "," + z );
				
				Object[] roomids = selected_rooms.keySet().toArray();
				
				// Loop over selected rooms and send save commands
				String tx = x + "," + y + "," + z;
				for( int i = 0; i < roomids.length; i++ ) {
					int roomid = ((Integer)roomids[ i ]).intValue();
					CheeseLab.dmesg( 4, "  > " + roomid );
					tx += "," + roomid;
				}
				CheeseMap.send_tx( 1, "move", tx );
				changed();
				// TODO: make sure if we are EDITING one of these that we update the room editor!
				// it might be easeier if the room tx handler checked to see if the room is in the
				// editor and just passed it on magically
			}
		}
	}
	
	public void set_tile_size( int width, int height ) {
		tile_width = width;
		tile_height = height;
		image_cache.clear();
		changed();
	}
	
	private void _start_timer( String section ) {
		timer_start = new Date();
		timer_section = section;
		System.err.println( "Timer started: " + section );
	}
	
	private void _end_timer() {
		timer_end = new Date();
		long diff = timer_end.getTime() - timer_start.getTime();
		double secs = (double)diff / 1000;
		System.err.println( timer_section + " took: " + diff + "ms (" + secs + " secs)" );
	}
	
	public void setEditTool(String toolname) {
		if ( toolname.equals( toolnames[TOOL_SELECT] ) ) {
			setCursor( tool_cursors[TOOL_SELECT] );
			tool_active = TOOL_SELECT;
		}else if ( toolname.equals( toolnames[TOOL_MERGE] ) ) {
			setCursor( tool_cursors[TOOL_MERGE] );
			tool_active = TOOL_MERGE;
		}else if ( toolname.equals( toolnames[TOOL_MOVE] ) ) {
			setCursor( tool_cursors[TOOL_MOVE] );
			tool_active = TOOL_MOVE;
		}else if ( toolname.equals( toolnames[TOOL_DELETE] ) ) {
			setCursor( tool_cursors[TOOL_DELETE] );
			tool_active = TOOL_DELETE;
		}else if ( toolname.equals( toolnames[TOOL_POINT] ) ) {
			setCursor( tool_cursors[TOOL_POINT] );
			tool_saved = getEditTool();
			tool_active = TOOL_POINT;
		}
	}
	
	public String getEditTool() {
		return toolnames[tool_active];
	}
	
	private void save_exit(int from, String dir, int to) {
		CheeseLab.dmesg( 3, "Save exit: " + from + " - " + dir + " -> " + to );

		// FIXME: This has an unsolvable race condition.
		//        basically what we do is we fetch the room info and then send it back.
		//        This is a waste of bandwidth and falls apart if some1 changes
		//        room data in the meantime. Also waiting for a ping freezes the GUI.
		//        Creating an atomic exit change command in broker would be a lot
		//        better idea
		//        eg. CheeseMap.send_tx( 1, "save_exit", from + "," + dir + "," + to );

		// Make sure it gets in the cache
		CheeseRoomData room = map.get_room( from, System.currentTimeMillis() );
		CheeseMap.send_tx(1, "ping", "set_edit_room");
		while ( !CheeseMap.checkping() );

		room.mtime = System.currentTimeMillis();
		if( dir.equalsIgnoreCase( "u" ) )
			room.exit_u = to;
		else if( dir.equalsIgnoreCase( "d" ) )
			room.exit_d = to;
		else if( dir.equalsIgnoreCase( "n" ) )
			room.exit_n = to;
		else if( dir.equalsIgnoreCase( "s" ) )
			room.exit_s = to;
		else if( dir.equalsIgnoreCase( "e" ) )
			room.exit_e = to;
		else if( dir.equalsIgnoreCase( "w" ) )
			room.exit_w = to;

		// Send save room command
		CheeseMap.send_tx( 1, "save", CheeseDecoder.encode_object(room) );

		changed();
	}
	
	private void save_overlay(int from, String ovrl, boolean en) {
	    Vector novrls = new Vector();
	    
	    CheeseLab.dmesg( 3, "Save overlay: " + from + " - " + ovrl + " -> " + en );
	    
		// Make sure it gets in the cache
		CheeseRoomData room = map.get_room( from, System.currentTimeMillis() );
		CheeseMap.send_tx(1, "ping", "set_edit_room");
		while ( !CheeseMap.checkping() );

		room.mtime = System.currentTimeMillis();
		// Copy all overlays except current one to vector 
		for( int i = 0; i < room.overlays.length; i++ ) {
		    String o = room.overlays[i];
		    if( !o.equals("") && !o.equals( ovrl ) )
		        novrls.add( o );
		}		
		// Add current overlay if requested so
		if( en )
		    novrls.add( ovrl );
		room.overlays = (String [])novrls.toArray( new String[0] ); 
		
		// Send save room command
		CheeseMap.send_tx( 1, "save", CheeseDecoder.encode_object(room) );

		changed();
	}
	
	/**
	 * Create custom cursor from a graphics file
	 * @param name  name to be looked up. The filename takes form: ui/ui_cursor_name.png
	 * @return      Cursor object with new cursor.
	 */
	Cursor createCursor(String name) {
	    // TODO: Scale cursor to system preferred size
		Toolkit tk = Toolkit.getDefaultToolkit();
		Image img = new ImageIcon( "ui/ui_cursor_" + name + ".png" ).getImage();
		return tk.createCustomCursor(img, new Point( 0,0 ), name);
	}
	
	/**
	 * Implement this class to add callback to point tool. This tool can be used
	 * to allow user to select just one room (to point a room).
	 */
	abstract public class PointToolCallback {
		abstract void pointedTo(Point p);
	}

	/**
	 * This class is used to implement Context Menu editor commands.
	 */
	class CtxEditorListener implements ActionListener {
		String sdata;

		int roomid;

		CtxEditorListener(String sdata, int roomid) {
			this.sdata = sdata;
			this.roomid = roomid;
		}

		public void actionPerformed(ActionEvent e) {
			String cmd = e.getActionCommand();
			if(cmd.equals( "set_exit_none" )) {
				save_exit( roomid, sdata, 0 );
			}else if( cmd.equals( "set_exit_unexplored" ) ) {
				save_exit( roomid, sdata, -1 );
			}else if( cmd.equals( "set_exit_dt" ) ) {
				save_exit( roomid, sdata, -2 );
			}else if( cmd.equals( "set_exit_from_map" ) ) {
				point_callback = new PointToolCallback() {
					void pointedTo(Point p) {
						Point3d p3d = local_to_global( p.x, p.y );

						Integer room = (Integer)mapdata.get( p3d.x + ","
								+ p3d.y + "," + p3d.z );
						if( room != null ) {
							save_exit( roomid, sdata, room.intValue() );
						}
					}
				};
				setEditTool( "point_tool" );
			}else if( cmd.equals( "set_overlay" )) {
			    JCheckBoxMenuItem src = (JCheckBoxMenuItem)e.getSource();
			    save_overlay( roomid, sdata, src.isSelected() );
			}
		}
	}
}
